import arrow


class MessageReader(object):
    """
    A tolerant reader for reading float, int, str, etc. from a message (dict).
    See 'tolerant reader' pattern.
    """
    def __init__(self, message):
        super().__init__()
        self.message = message

    def bool_value(self, keys, default=False):
        value = self.get_value(self.message, keys)
        if isinstance(value, bool):
            return value

        for m in ["true", True, "1", 1]:
            if value == m and type(value) == type(m):
                return True

        for m in ["false", False, "0", 0]:
            if value == m and type(value) == type(m):
                return False

        return default

    def dict_value(self, keys, default={}):
        value = self.get_value(self.message, keys)
        if isinstance(value, dict):
            return value
        else:
            return default

    def list_value(self, keys, default=[]):
        value = self.get_value(self.message, keys)
        if isinstance(value, list):
            return value
        else:
            return default

    def string_value(self, keys, default=None):
        value = self.get_value(self.message, path=keys, default=default)
        if isinstance(value, str):
            return value
        elif value is None:
            return default
        return str(value)

    def entity_id_value(self, keys, class_, default=None):
        value = self.get_value(self.message, path=keys, default=default)
        if isinstance(value, str):
            return class_(value)
        elif value is None:
            return default
        return None

    def lang_value(self, keys, default=None):
        value = self.get_value(self.message, path=keys, default=default)
        _map = {
            'en': 'en',
            'eng': 'en',
            'sv': 'sv',
            'swe': 'sv',
        }
        if isinstance(value, str):
            if value in _map:
                return _map[value]
            return value
        return default

    def date_value(self, keys):
        value = self.get_value(self.message, keys)
        if not isinstance(value, str):
            return None
        else:
            date = None
            try:
                date = arrow.get(value)
            except:
                pass
            return date

    def date_value_iso8601(self, keys):
        date = self.date_value(keys=keys)
        if date:
            return date.format("YYYY-MM-DDTHH:mm:ss.SSSSSSZZ")
        return None

    def int_value(self, keys):
        value = self.get_value(self.message, keys)
        if isinstance(value, int):
            return value
        else:
            return int(value) if isinstance(value, str) and value.isdigit() else None

    def float_value(self, keys):
        value = self.get_value(self.message, keys)
        if isinstance(value, float):
            return value
        elif isinstance(value, int):
            return float(value)
        else:
            return float(value) if isinstance(value, str) and value.replace('.','',1).isdigit() else None

    def has_value(self, keys):
        value = self.get_value(self.message, keys, default=None)
        return value is not None

    def get_value(self, the_dict, path, default=""):
        keys = path.split('.')
        if the_dict != None and keys[0] in the_dict:
            if len(keys) == 1:
                return the_dict[keys[0]]
            else:
                return self.get_value(the_dict[keys[0]], ".".join(keys[1:]), default)
        else:
            return default



import asyncio
import json

import asyncpg

from ddd.infrastructure.db_service import DbService


class PostgresDbService(DbService):
    """
    A postgres db service.
    """
    def __init__(
        self,
        dsn,
        log_service,
        min_size=20,
        max_size=20,
    ):
        super().__init__(log_service=log_service)
        self.dsn = dsn
        self.min_size = min_size
        self.max_size = max_size

    async def _create_conn_pool(self):
        self.log_service.info(
            f"..creating connection pool "
            f"(min: {self.min_size}, max: {self.max_size}).."
        )

        created = False
        error = None
        backoff = 6
        retries = 0
        max_retries = 5

        async def init_conn(conn):
            await conn.set_type_codec(
                'json',
                encoder=json.dumps,
                decoder=json.loads,
                schema='pg_catalog'
            )

        while (not created) and retries < max_retries:
            try:
                pool = \
                    await asyncpg.create_pool(
                        dsn=self.dsn,
                        min_size=self.min_size,
                        max_size=self.max_size,
                        init=init_conn
                    )
                created = True
            except Exception as e:
                self.log_service.error(
                    (
                        "Failed to create connection pool of postgres db "
                        "at: {} (retrying in {} secs..), exception: {}"
                    ).
                    format(
                        self.dsn,
                        backoff,
                        str(e)
                    )
                )
                error = str(e)
                retries += 1
                await asyncio.sleep(backoff)

        if not created:
            raise Exception(
                (
                    "App failed to create connection pool of db at: {} "
                    "after {} retries, error: {}."
                ).
                format(
                    self.dsn,
                    retries,
                    error
                )
            )

        self.conn_pool = pool



class MemoryEventRepository(object):
    def __init__(self):
        super(MemoryEventRepository, self).__init__()
        self.events = {
            'domain': {},
            'integration': {},
        }

    async def delete_all(self, action_id, event_type):
        self.events[event_type].pop(action_id, None)

    async def get_unpublished(self, action_id, event_type):
        if action_id in self.events[event_type]:
            return self.events[event_type][action_id]
        return []

    async def save_all(self, action_id, events, event_type):
        self.events[event_type][action_id] = events



"""
Collection of utilities until we find a better place to put them.
"""
import copy

import arrow
import base64
import hashlib
import json
import logging
import os

from addict import Dict
from dotenv import find_dotenv, load_dotenv
from time import time

from ddd.domain.building_block import BuildingBlock
from ddd.domain.entity import Entity

logger = logging.getLogger(__file__)


# Functions

def read_config():
    """
    Read config (from environment variables).

    Returns:
        dict - the config dict
    """
    truthy = ["True", "true", "1", 1, True]

    config = Dict({
        'env': os.getenv("ENV"),
        'debug': os.getenv("DEBUG"),
        'max_concurrent_actions':
            int(os.getenv("MAX_CONCURRENT_ACTIONS", "10")),
        'top_level_package_name': os.getenv("TOP_LEVEL_PACKAGE_NAME"),
        'http': {
            'debug': os.getenv("HTTP_DEBUG") in truthy,
            'port': os.getenv("HTTP_PORT"),
        },
        'database': {
            'type': os.getenv("DATABASE_TYPE"),
            'postgres': {
                'dsn': os.getenv("POSTGRES_DSN"),
            },
        },
        'auth': {
            'full_access_token': os.getenv("AUTH_FULL_ACCESS_TOKEN"),
        },
        'pubsub': {
            'domain': {
                'provider': os.getenv("DOMAIN_PUBSUB_PROVIDER"),
                'topic': os.getenv("DOMAIN_PUBSUB_TOPIC"),
                'group': os.getenv("DOMAIN_PUBSUB_GROUP"),
            },
            'interchange': {
                'provider': os.getenv("INTERCHANGE_PUBSUB_PROVIDER"),
                'topic': os.getenv("INTERCHANGE_PUBSUB_TOPIC"),
                'group': os.getenv("INTERCHANGE_PUBSUB_GROUP"),
            },
            'kafka': {
                'bootstrap_servers': os.getenv("KAFKA_BOOTSTRAP_SERVERS"),
            },
            'azure': {
                'namespace': os.getenv("AZURE_EVENT_HUBS_NAMESPACE"),
                'namespace_conn_string':
                    os.getenv("AZURE_EVENT_HUBS_NAMESPACE_CONN_STRING"),
                'checkpoint_store_conn_string':
                    os.getenv("AZURE_EVENT_HUBS_CHECKPOINT_STORE_CONN_STRING"),
                'blob_container_name':
                    os.getenv("AZURE_EVENT_HUBS_BLOB_CONTAINER_NAME"),
            },
        },
        'jobs': {
            'scheduler': {
                'type': os.getenv("JOBS_SCHEDULER_TYPE"),
                'dsn': os.getenv("JOBS_SCHEDULER_DSN"),
            },
        },
        'slack': {
            'token': os.getenv("SLACK_TOKEN"),
        },
        'log': {
            'enabled': True,
            'slack': {
                'enabled': os.getenv("LOG_SLACK_ENABLED") in truthy,
                'errors': {
                    'channel': os.getenv("LOG_SLACK_CHANNEL_ERRORS"),
                }
            },
            'kibana': {
                'enabled': os.getenv("LOG_KIBANA_ENABLED") in truthy,
                'url': os.getenv("LOG_KIBANA_URL"),
                'username': os.getenv("LOG_KIBANA_USERNAME"),
                'password': os.getenv("LOG_KIBANA_PASSWORD"),
            }
        },
        'base_url': os.getenv('BASE_URL', 'empty-base-url'),
        'api': {

        },
        'tasks': {
            'username': os.getenv('TASK_USERNAME'),
            'password': os.getenv('TASK_PASSWORD'),
        }
    })

    return config

def load_env_file():
    """
    Load environment variables from env file.
    """
    filename = os.getenv("ENV_FILE", "env")

    loaded = \
        load_dotenv(
            find_dotenv(filename=filename)
        )

    if not loaded:
        raise Exception(f"Couldn't load env file with name: '{filename}'")

def sri(path):
    """
    Calculate the SRI checksum of a file at 'path'.
    """
    with open(path, "r") as f:
        input = f.read()

        if isinstance(input, str):
            # This is so we can handle strings as input in Python 3
            input = input.encode()

        hash = hashlib.sha256(input).digest()
        hash_base64 = base64.b64encode(hash).decode()

        return 'sha256-{}'.format(hash_base64)

def dot_pop(the_dict, path):
    """
    Like native dict.pop(), but supports dot-based paths.
    Ex. dot_pop(obj, 'my.dot.path.to.field')
    """
    keys = path.split('.')
    if the_dict != None and keys[0] in the_dict:
        if len(keys) == 1:
            the_dict.pop(keys[0])
        else:
            dot_pop(the_dict[keys[0]], ".".join(keys[1:]))

def flatten(the_dict):
    """
    Flatten 'the_dict' by replacing any Entity with it's
    __dict__ version.

    TODO: Recurse..
    """
    for key, value in the_dict.items():
        if isinstance(the_dict[key], Entity):
            the_dict[key] = the_dict[key].__dict__
    return the_dict

def get_for_compare(obj, ignore_fields=None):
    """
    Get representation for comparision.

    - Deep copy
    - Flatten entities
    - Pop 'ignore_fields'
    """
    ignore_fields = ignore_fields if ignore_fields is not None else []

    if type(obj) in [str, int, float]:
        return obj
    elif type(obj) == list:
        return [get_for_compare(i) for i in obj]

    obj = copy.deepcopy(obj.__dict__)

    for key, value in obj.items():
        if isinstance(value, BuildingBlock):
            obj[key] = get_for_compare(value)
        elif isinstance(value, list):
            for n, i in enumerate(value):
                value[n] = get_for_compare(i)

    obj = flatten(obj)

    for field in ignore_fields:
        dot_pop(obj, field)

    return obj

# Dates

def date_or_none(string):
    if string is None:
        return None
    return arrow.get(string)

def iso_or_none(date):
    if date is None:
        return None
    return date.isoformat()

def str_or_none(string):
    if string is None:
        return None
    return str(string)

def dates_intersection(
    first_start,
    first_end,
    second_start,
    second_end,
    timezone,
):
    if None in [first_start, first_end, second_start, second_end]:
        return None

    first_start.to(timezone)
    first_end.to(timezone)
    second_start.to(timezone)
    second_end.to(timezone)

    start = None
    end = None

    if first_start > second_start:
        start = first_start
    else:
        start = second_start

    if first_end < second_end:
        end = first_end
    else:
        end = second_end

    if start > end:
        return None

    return [start, end]

# Classes

class catch_time(object):
    """
    Context manager that measures the execution time.
    """
    def __enter__(self):
        self.start = time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.execution_time = time() - self.start


class MockResponse(object):
    """
    Used to mock responses returned from a requests package request.
    """
    def __init__(
        self,
        status_code,
        content,
        content_type,
        headers=None,
        url="mock-url"
    ):
        headers = headers if headers is not None else {}

        self.status_code = status_code
        self.content_type = content_type
        self.headers = headers
        self.url = url

        # Make sure content is a byte string
        if type(content) is not str:
            content = json.dumps(content)

        self._content = content.encode('utf-8')

    async def content(self):
        return self._content

    async def json(self):
        return json.loads(self._content)

    async def text(self):
        return self._content



import aiosonic
import argparse

from abc import ABCMeta, abstractmethod


class Task(object, metaclass=ABCMeta):
    """
    A base class for tasks.
    """
    def __init__(self, config, deps_mgr, args_str, makes_requests=False):
        super().__init__()

        self.config = config
        self.deps_mgr = deps_mgr
        self.args_str = args_str
        self.makes_requests = makes_requests

        self.log_service = deps_mgr.get_log_service()

        self._create_args_parser()
        self.add_args(parser=self.parser)
        self._parse_args()

        self.client = None
        self.token = None
        self.pool = None

    # Interface

    @abstractmethod
    def add_args(self, parser):
        """
        Override this method in subclasses.
        """
        pass

    @abstractmethod
    async def run(self):
        """
        Override this method in subclasses.
        """
        pass

    # Init

    def _create_args_parser(self):
        self.parser = argparse.ArgumentParser()

    def _parse_args(self):
        args = []

        if len(self.args_str) > 0:
            args = self.args_str.split(" ")

        self.args = self.parser.parse_args(args=args)

    # Run

    async def _run(self):
        if "-h" in self.args or "--help" in self.args:
            self.parser.print_help()
        else:
            await self.initialize()
            await self.run()

    async def initialize(self):
        """
        Override this method if you need to do some initialization
        in your subclass before run() is called,
        (for example initiate state variables).
        """
        # Adapters
        await self._create_secondary_adapters()

        # Client
        if self.makes_requests:
            self.client = aiosonic.HTTPClient()
            await self._login()

    # HTTP

    async def _create_secondary_adapters(self):
        """
        Create secondary adapters.
        """
        # Scheduler adapter
        self.scheduler_adapter = self.deps_mgr.get_scheduler_adapter()

        await self.scheduler_adapter.start()

    async def _login(self):
        raise NotImplementedError()

    # Helpers

    def _non_empty_string(self, string):
        """
        Custom type for argparse
        """
        if not string:
            raise ValueError("Must not be empty string")
        return string



from ddd.infrastructure.db_service import DbService
from ddd.infrastructure.memory.memory_pool import MemoryPool


class MemoryDbService(DbService):
    """
    A memory db service.
    """
    def __init__(
        self,
        log_service,
    ):
        super().__init__(log_service=log_service)

    async def _create_conn_pool(self):
        self.log_service.info("Creating memory db pool")
        self.conn_pool = MemoryPool(max=80)



import arrow
import dateutil

from abc import abstractmethod


class Repository(object):

    def __init__(
        self,
        log_service,
        aggregate_cls,
        translator_cls,
    ):
        super().__init__()
        self.log_service = log_service
        self.aggregate_cls = aggregate_cls
        self.translator_cls = translator_cls

    # Init

    @abstractmethod
    async def connect(self):
        pass

    @abstractmethod
    async def assert_table(self):
        pass

    @abstractmethod
    async def disconnect(self):
        pass

    # Control

    async def start(self):
        await self.connect()
        await self.assert_table()

    async def stop(self):
        await self.disconnect()

    # Operations

    async def _get(self, aggregate_id):
        record = await self._get_record(aggregate_id=aggregate_id)

        aggregate = None

        if record:
            aggregate = self.aggregate_from_record(record)

        return aggregate

    async def get_all(self):
        records = await self._get_all_records()
        aggregates = [self.aggregate_from_record(record) for record in records]
        return aggregates

    async def get_all_not_on_latest_version(self):
        records = await self._get_all_records_not_on_latest_version()
        records = self._migrate(records, self.aggregate_cls)
        aggregates = [self.aggregate_from_record(record) for record in records]
        return aggregates

    @abstractmethod
    async def _get_record(self, aggregate_id):
        pass

    @abstractmethod
    async def _get_all_records(self):
        pass

    @abstractmethod
    async def _get_all_records_not_on_latest_version(self):
        pass

    @abstractmethod
    async def next_identity(self):
        pass

    async def _save(self, aggregate_id, aggregate):
        await self._save_record(aggregate_id, self._serialize(aggregate))

    @abstractmethod
    async def _save_record(self, aggregate_id, data):
        pass

    # Dates

    def datetime_from_iso8601(self, iso8601, tz_if_missing="Europe/Stockholm"):
        if iso8601 is None:
            return None
        
        # Contains timezone?
        date = dateutil.parser.parse(iso8601)
        
        if date.tzinfo is None:
            # ..no, use default
            date = arrow.get(
                iso8601,
                "YYYY-MM-DDTHH:mm:ss.SSSSSSZZ",
                tzinfo=tz_if_missing
            )
            
        else:
            # ..yes, use it
            date = arrow.get(
                iso8601,
                "YYYY-MM-DDTHH:mm:ss.SSSSSSZZ"
            )
        
        return date

    def now(self, tz="Europe/Stockholm"):
        now = arrow.utcnow().to(tz)
        return now

    def now_iso8601(self, tz="Europe/Stockholm"):
        string = self.now(tz).format("YYYY-MM-DDTHH:mm:ss.SSSSSSZZ")
        return string

    # Serialization / Deserialization

    def aggregate_from_record(self, record):
        aggregate = self.translator_cls.from_record(record)
        return aggregate

    def _serialize(self, aggregate):
        """
        Serializes an aggregate by turning it into a dict.
        """
        data = self.translator_cls(aggregate).to_domain()
        return data

    # Migration

    def _migrate(self, records, cls):
        """
        Migrate the record to latest version.
        """
        migrated = []

        if len(records):
            if not hasattr(cls, "VERSION"):
                raise Exception(
                    f"Couldn't migrate entity. "
                    f"The '{cls}' class is missing 'VERSION'."
                )

        for r in records:

            versions = \
                self._get_versions_to_migrate_to(
                    record=r,
                    cls=cls
                )

            if versions.stop > 0:
                prev = versions.start - 1

                for next_ in range(versions.start, versions.stop + 1):
                    func_name = f"_migrate_v{prev}_to_v{next_}"
                    func = getattr(self, func_name)

                    if not func:
                        raise Exception(
                            f"Couldn't migrate entity. "
                            f"The '{cls}' class is missing the "
                            f"'{func_name}' function"
                        )

                    # Log
                    self.log_service.debug(
                        f"Migrating '{self.aggregate_cls.__name__}' "
                        f"from 'v{prev}' to 'v{next_}'."
                    )

                    # Migrate record to next version
                    r = func(record=r)

                    # Inc. version
                    r['data']['version'] = str(next_)

                    prev = next_

            migrated.append(r)

        return migrated

    def _get_versions_to_migrate_to(self, record, cls):
        """
        Get a range of consecutive versions to migrate to.
        """
        latest = int(cls.VERSION)
        current = self._get_version_from_record(record)

        if latest == current:
            versions = range(0, 0)
        else:
            versions = range(current + 1, latest)

        return versions

    def _get_version_from_record(self, record):
        version = int(record['data']['version'])
        return version



import importlib
import inspect
from pathlib import Path
import pkgutil

from ddd.utils.tasks.task import Task as BaseTask


class Task(BaseTask):

    def __init__(self, config, deps_mgr, args_str):
        super().__init__(
            config=config,
            deps_mgr=deps_mgr,
            args_str=args_str,
            makes_requests=False,
        )

    def add_args(self, parser):
        pass

    async def run(self):

        # Migrate
        results = {}

        repositories = self._get_repositories()

        for repository in repositories:

            name = self._get_name_of_root(repository)

            if name not in results:
                results[name] = {
                    'count': 0,
                }

            aggregates = await repository.get_all_not_on_latest_version()

            for aggregate in aggregates:
                await repository.save(aggregate)
                results[name]['count'] += 1

        # Print results
        rows = []

        for name, data in results.items():
            rows.append(
                f"* Migrated '{data['count']}' {name.lower()}s"
            )

        self.log_service.info(
            "---------------------------------------\nRESULT:\n\n"
            "{}\n"
            "---------------------------------------".
            format(
                "\n".join(rows)
            )
        )

    def _get_repositories(self):
        """
        Get all repositories.
        """
        repositories = []

        modules = self._get_modules()

        for module, path in modules.items():

            spec = importlib.util.spec_from_file_location(module, f"{path}.py")
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)

            module_classes = inspect.getmembers(mod, inspect.isclass)

            ignored_classes = [
                "MemoryRepository", "PostgresRepository",
                "MemoryEventRepository", "PostgresEventRepository",
                "Repository",
            ]

            for name, cls in module_classes:
                if self._camel_to_snake(name) == cls.__module__:
                    if "Repository" in name:
                        if name not in ignored_classes:
                            _type = self.config.database.type
                            if _type.capitalize() in name:
                                func_name = name
                                func_name = \
                                    func_name.replace(_type.capitalize(), "")
                                func_name = self._camel_to_snake(func_name)
                                func_name = f"get_{func_name}"
                                repository = getattr(self.deps_mgr, func_name)()
                                repositories.append(repository)

        return repositories

    def _get_modules(self, path=None):
        """
        Get all modules under the '<package_name>/repositories' folder.
        """
        if path is None:
            path = \
                Path(__file__).parents[3] / \
                f"{self.config.top_level_package_name}/repositories"

        modules = {}

        for loader, module_name, is_pkg in pkgutil.walk_packages([path]):

            if is_pkg:
                modules.update(
                    self._get_modules(
                        path=path / module_name
                    )
                )
            else:
                modules[module_name] = str(path / module_name)

        return modules

    def _get_name_of_root(self, repository):
        return repository.aggregate_cls.__name__

    def _camel_to_snake(self, string):
        """
        Get camel case version of snake case 'string'
        """
        return ''.join(
            ['_' + c.lower() if c.isupper() else c for c in string]
        ).lstrip('_')



import copy
import json
import logging
import sys

from addict import Dict

from ddd.infrastructure.infrastructure_service import InfrastructureService

from logging.config import dictConfig

from cmreslogging.handlers import CMRESHandler


class LogService(InfrastructureService):
    """
    A log service for logging.
    """
    def __init__(
        self,
        kibana_config,
        slack_config,
        enabled=True,
    ):
        super().__init__(log_service="dummy")

        self.kibana_config = Dict(kibana_config)
        self.slack_config = Dict(slack_config)
        self.enabled = enabled

        # Create the logger
        self._set_modules_logger_levels()
        self._create_logger()

    async def start(self):
        pass

    async def stop(self):
        pass

    def _set_modules_logger_levels(self):
        """
        Sets log level of used modules.

        Effectively remove/add unwanted/wanted logs of third-party packages
        from/to console output.
        """
        logging.getLogger("aiokafka").setLevel(logging.WARNING)
        logging.getLogger("azure").setLevel(logging.WARNING)
        logging.getLogger("elasticsearch").setLevel(logging.WARNING)
        logging.getLogger("uamqp").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)

    def info(self, message, extra=None, exc_info=False):
        """
        Log an info level message.
        """
        if self.enabled:
            LogService._do_log(
                logger=self.logger,
                level="info",
                message=message,
                extra=extra,
                exc_info=exc_info,
            )

    def warning(self, message, extra=None, exc_info=False):
        """
        Log an warning level message.
        """
        if self.enabled:
            LogService._do_log(
                logger=self.logger,
                level="warning",
                message=message,
                extra=extra,
                exc_info=exc_info,
            )

    def error(self, message, extra=None, exc_info=False):
        """
        Log an error level message.
        """
        if self.enabled:
            LogService._do_log(
                logger=self.logger,
                level="error",
                message=message,
                extra=extra,
                exc_info=exc_info,
            )

    def debug(self, message, extra=None, exc_info=False):
        """
        Log an debug level message.
        """
        if self.enabled:
            LogService._do_log(
                logger=self.logger,
                level="debug",
                message=message,
                extra=extra,
                exc_info=exc_info,
            )

    async def log_request(self, message, method, url, params):
        """
        Convenience method to log the request to an api.
        """
        if self.enabled:

            params = copy.deepcopy(params)

            if 'token' in params:
                params['token'] = "(hidden)"

            if 'password' in params:
                params['password'] = "(hidden)"

            self.debug(
                message,
                extra={
                    'method': method,
                    'url': url,
                    'params': params,
                }
            )

    async def log_response(self, message, method, url, params, response):
        """
        Convenience method to log the response of an api request.
        """
        if self.enabled:
            content = response.content
            params = copy.deepcopy(params)

            if response.headers.get('content-type') == 'application/json':
                content = json.loads(await response.content())
            else:
                content = await response.content()

            if 'token' in params:
                params['token'] = "(hidden)"

            if 'password' in params:
                params['password'] = "(hidden)"

            self.debug(
                message,
                extra={
                    'method': method,
                    'url': url,
                    'params': params,
                    'status_code': response.status_code,
                    'content_type': response.headers.get('content-type'),
                    'response': content,
                }
            )

    # Helpers

    @classmethod
    def _do_log(
        cls,
        logger,
        level,
        message,
        extra,
        exc_info,
    ):
        """
        Helper to do the actual logging.

        Returns:
            Nothing if logging succeeds.

        Raises:
            Exception if logging fails.
        """
        e = None

        if extra is not None and extra != {}:
            e = {'extra': extra}

        getattr(logger, level)(
            message,
            extra=e,
            exc_info=exc_info,
        )

    def _create_logger(self):

        # TODO: Merge this into logging_config below:
        logging.basicConfig(
            format=
            "[%(asctime)s] [%(levelname)8s] --- "
            "%(message)s (%(filename)s:%(lineno)s)",
            stream=sys.stdout,
            level=logging.DEBUG
        )

        logging_config = dict(
            disable_existing_loggers=False,
            version=1,
            formatters={
                'f': {
                    'format': (
                        "%(asctime)s %(name)-12s %(levelname)-8s "
                        "%(message)s"
                    )
                }
            },
            handlers={
                'console': {
                    'class': "logging.StreamHandler",
                    'formatter': None,
                    'level': logging.INFO,
                    'stream': "ext://sys.stdout"
                },
                'kibana': {
                    'class': "cmreslogging.handlers.CMRESHandler",
                    'formatter': None,
                    'level': logging.DEBUG,
                    'hosts': f"{self.kibana_config.url}",
                    'auth_type': CMRESHandler.AuthType.BASIC_AUTH,
                    'auth_details': (
                        self.kibana_config.username,
                        self.kibana_config.password,
                    ),
                    'use_ssl': "443" in self.kibana_config.url
                    if self.kibana_config.url else False,
                    'verify_ssl': "443" in self.kibana_config.url
                    if self.kibana_config.url else False,
                    'es_index_name': "logstash",
                    'es_additional_fields': {
                        'source': self.kibana_config.source,
                    },
                    'raise_on_indexing_exceptions': False
                },
                'slack-error': {
                    'level': 'ERROR',
                    'api_key': self.slack_config.token,
                    'username': self.slack_config.bot_username,
                    'class': 'slacker_log_handler.SlackerLogHandler',
                    'channel': f"#{self.slack_config.errors.channel}",
                },
            },
            loggers={
                'ddd-for-python': {
                    'level': logging.DEBUG,
                    'propagate': True,
                    'filters': [],
                    'handlers': [],
                }
            }
        )

        # Add handlers
        handlers = ['console']

        if self.kibana_config.enabled:
            handlers.append('kibana')

        if self.slack_config.enabled:
            handlers.append('slack-error')

        logging_config['loggers']['ddd-for-python']['handlers'] = handlers

        dictConfig(logging_config)

        self.logger = logging.getLogger("ddd-for-python")



from ddd.application.application_service import ApplicationService
from ddd.application.action import action


class DummyApplicationService(ApplicationService):
    """
    This is the dummy context's application service.
    """
    def __init__(
        self,
        db_service,
        domain_adapter,
        domain_publisher,
        event_repository,
        interchange_adapter,
        interchange_publisher,
        job_service,
        job_adapter,
        log_service,
        scheduler_adapter,
        dummy_repository,
        max_concurrent_actions,
        loop=None,
    ):
        """
        Initialize the application service.
        """
        super().__init__(
            db_service=db_service,
            domain_adapter=domain_adapter,
            domain_publisher=domain_publisher,
            event_repository=event_repository,
            interchange_adapter=interchange_adapter,
            interchange_publisher=interchange_publisher,
            job_service=job_service,
            job_adapter=job_adapter,
            log_service=log_service,
            scheduler_adapter=scheduler_adapter,
            max_concurrent_actions=max_concurrent_actions,
            loop=loop,
        )

        # Dependencies
        self.dummy_repository = dummy_repository

        # ..categorize
        self.domain_services.extend([

        ])

        self.primary_adapters.extend([

        ])

        self.secondary_adapters.extend([
            dummy_repository,
        ])

    # Actions

    @action
    async def do_something(self, command, corr_ids=None):
        """
        Do something.
        """
        corr_ids = corr_ids if corr_ids is not None else []

        raise NotImplementedError()



import asyncio
import signal


class Container(object):
    """
    A container for hosting an application service.
    """
    def __init__(self, app_service, log_service, loop=None):
        super().__init__()

        # Deps
        self.app_service = app_service
        self.log_service = log_service
        self.loop = loop if loop else asyncio.get_event_loop()

        # Vars
        self.stop_sem = None

        # Docker
        self._create_stop_semaphore()
        self._subscribe_to_signals()

    # Docker

    def _create_stop_semaphore(self):
        """
        Create the stop semaphore that we will block on after application
        service has been started.

        Acquiring the semaphore means the container has been ordered to stop.
        """
        self.stop_sem = asyncio.Event()

    def _subscribe_to_signals(self):
        """
        Subscribe to SIGINT (ctrl + c) and SIGTERM (docker stop).
        """
        for signal_name in ('SIGINT', 'SIGTERM'):
            self.loop.add_signal_handler(
                getattr(signal, signal_name),
                lambda: self.loop.create_task(self._handle_signal(signal_name))
            )

    async def _handle_signal(self, signal_name):
        """
        Handler for process signals.
        """
        self.log_service.debug(f"Received system signal: '{signal_name}'.")

        if signal_name in ['SIGTERM', 'SIGKILL']:
            self.stop_sem.set()

    # Run

    async def run(self):
        """
        Run the app, (called on container start).
        """
        await self.app_service.start()
        await self.wait_for_stop()
        await self.stop()

    async def wait_for_stop(self):
        """
        Block waiting for termination/stop signals.
        """
        self.log_service.info("Waiting for stop/term signals..")
        await self.stop_sem.wait()

    async def stop(self):
        self.log_service.info("Stopping app..")
        await self.app_service.stop()



from ddd.repositories.memory.memory_repository import MemoryRepository

from ddd.domain.dummy.dummy import Dummy
from ddd.domain.dummy.dummy_repository import DummyRepository
from ddd.domain.dummy.dummy_translator import DummyTranslator


class MemoryDummyRepository(DummyRepository, MemoryRepository):

    def __init__(self, log_service):
        MemoryRepository.__init__(
            self,
            log_service=log_service,
            aggregate_name="dummy",
            aggregate_cls=Dummy,
            translator_cls=DummyTranslator,
        )

    async def get(self, dummy_id):
        return await super()._get(dummy_id)

    async def save(self, dummy):
        await super()._save(
            aggregate_id=dummy.dummy_id,
            aggregate=dummy,
        )

    async def delete(self, dummy):
        await self._delete(dummy.dummy_id)



import asyncio
import copy
import uuid

from ddd.application.domain_registry import DomainRegistry
from ddd.domain.exceptions import DomainException


def action(func):
    """
    A decorator for the application service's "action" methods.

    Calls the wrapped method inside a transaction. Then saves any
    published events to the database.

    If transaction is committed (not rolled back), the events are
    actually published and then deleted from the database.

    The flow described above makes sure state is persisted and events
    are published atomically (either both happens or nothing). This
    assures us the overall system state is consistent.
    """
    async def func_wrapper(self, *args, **kwargs):
        async with self.action_sem:
            task_id = id(asyncio.current_task())
            action_id = uuid.uuid4().hex
            domain_exception = None
            result = None

            # Get dependencies
            domain_publisher = self.domain_publisher
            interchange_publisher = self.interchange_publisher
            event_repository = self.event_repository

            # Log
            command = None

            if 'command' in kwargs:
                command = copy.deepcopy(kwargs.get('command')).__dict__
                if 'token' in command:
                    command['token'] = "(hidden)"
                if 'password' in command:
                    command['password'] = "(hidden)"
                if 'new_password' in command:
                    command['new_password'] = "(hidden)"

            if 'corr_ids' in kwargs and kwargs['corr_ids']:
                corr_ids = command['corr_ids']
            else:
                corr_ids = []

            self.log_service.info(
                f"[ ACTION ]: {func.__name__}",
                extra={
                    'extra': {
                        'command': command,
                    },
                    'corr_ids': corr_ids,
                }
            )

            try:
                async with self.db_service.conn_pool.acquire() as conn:
                    DomainRegistry.get_instance().task_db_conns[task_id] = conn

                    # BEGIN TRANSACTION
                    async with conn.transaction():

                        # Run action
                        try:
                            result = await func(self, *args, **kwargs)
                        except Exception as e:
                            if not isinstance(e, DomainException):
                                raise Exception(
                                    f"Failed to run action, "
                                    f"exception: {str(e)}"
                                )
                            domain_exception = e

                        # Save domain events
                        try:
                            published = domain_publisher.get_published()
                            if len(published):
                                await event_repository.save_all(
                                    action_id=action_id,
                                    events=published,
                                    event_type="domain"
                                )
                            domain_publisher.clear_published()
                        except Exception as e:
                            raise Exception(
                                f"Failed to persist domain events, "
                                f"exception: {str(e)}"
                            )

                        # Save integration events
                        try:
                            published = interchange_publisher.get_published()
                            if len(published):
                                await event_repository.save_all(
                                    action_id=action_id,
                                    events=published,
                                    event_type="integration"
                                )
                            interchange_publisher.clear_published()
                        except Exception as e:
                            raise Exception(
                                f"Failed to persist interchange events, "
                                f"exception: {str(e)}"
                            )
                    # END TRANSACTION

                    # Flush domain events
                    try:
                        events = await event_repository.get_unpublished(
                            action_id=action_id,
                            event_type="domain"
                        )
                        for event in events:
                            self.log_service.debug(
                                f"Flushing domain event: '{event.name}'"
                            )
                            await domain_publisher.flush(event)

                        await event_repository.delete_all(
                            action_id=action_id,
                            event_type="domain"
                        )

                    except Exception as e:
                        raise Exception(
                            f"Failed to publish domain events, "
                            f"exception: {str(e)}"
                        )

                    # Flush integration events
                    try:
                        events = \
                            await event_repository.get_unpublished(
                                action_id=action_id,
                                event_type="integration"
                            )
                        for event in events:
                            self.log_service.debug(
                                f"Flushing interchange event: '{event.name}'"
                            )
                            await interchange_publisher.flush(event)

                        await \
                            event_repository.delete_all(
                                action_id=action_id,
                                event_type="integration"
                            )

                    except Exception as e:
                        raise Exception(
                            f"Failed to publish integration events "
                            f"with exception: {str(e)}"
                        )

            finally:

                # Remove db conn reference
                if task_id in DomainRegistry.get_instance().task_db_conns:
                    del DomainRegistry.get_instance().task_db_conns[task_id]

            # Raise action exception, if occured
            if domain_exception:
                raise domain_exception

            return result

    return func_wrapper



from ddd.infrastructure.infrastructure_service import InfrastructureService


class EmailService(InfrastructureService):
    """
    An email infrastructure service used to send emails.
    """
    def __init__(
        self,
        logger=None,
    ):
        super().__init__(logger=logger)

    async def send(
        self,
        from_,
        to_,
        subject,
        content,
    ):
        raise NotImplementedError()



import asyncio
import json

from azure.eventhub.aio import EventHubProducerClient
from azure.eventhub import EventData

from ddd.domain.event_publisher import EventPublisher


class AzureEventPublisher(EventPublisher):
    """
    A azure event publisher.
    """
    def __init__(
        self,
        namespace,
        namespace_conn_string,
        topic,
        group,
        log_service,
        keep_flushed_copies=False,
    ):
        super().__init__(
            log_service=log_service,
            keep_flushed_copies=keep_flushed_copies
        )
        self.namespace = namespace
        self.namespace_conn_string = namespace_conn_string
        self.topic = topic
        self.group = group

    # Flush

    async def flush(self, event):
        """
        Actually publishes the event,
        (called after request is done in the 'action' decorator).
        """
        task_id = self._get_loop_task_id()

        if not self.producer:
            raise Exception(
                (
                    "Azure event publisher couldn't publish because "
                    "producer has not been created."
                )
            )

        # Serialize
        try:
            json_data = json.dumps(event.serialize()).encode()
        except TypeError:
            self.log_service.error(
                "Failed to flush an event! The event couldn't be serialized, "
                "(check log details for event data).",
                extra={
                    'event_data': event.serialize(),
                }
            )
            raise

        # Without specifying partition_id or partition_key
        # the events will be distributed to available partitions via
        # round-robin.
        event_data_batch = await self.producer.create_batch()

        event_data_batch.add(
            EventData(
                json_data
            )
        )
        await self.producer.send_batch(event_data_batch)

        if self.keep_flushed_copies:
            if task_id not in self.flushed:
                self.flushed[task_id] = []

            self.flushed[task_id].append(event)

    # Control

    async def start(self):
        await self.create_producer()

    async def stop(self):
        if not self.producer:
            self.log_service.warning(
                (
                    "Azure event publisher was instructed to stop "
                    "but it doesn't seem to be started (no producer)."
                )
            )
        else:
            await self.producer.close()

    # Publishing

    async def create_producer(self):
        self.producer = \
            EventHubProducerClient.from_connection_string(
                conn_str=self.namespace_conn_string,
                eventhub_name=self.topic,
            )
        await asyncio.sleep(0.01)



class EntityIdTranslator(object):
    def __init__(self, entity_id):
        super().__init__()
        self.entity_id = entity_id

    @classmethod
    def from_domain(cls, identity, entity_class):
        entity_id = entity_class(identity=identity)
        return entity_id



import asyncio

from abc import ABCMeta

from ddd.adapters.event.azure.azure_event_adapter import AzureEventAdapter
from ddd.adapters.event.kafka.kafka_event_adapter import KafkaEventAdapter
from ddd.adapters.event.memory.memory_event_adapter import MemoryEventAdapter
from ddd.adapters.job.ap_scheduler_adapter import ApSchedulerAdapter
from ddd.adapters.job.job_adapter import JobAdapter
from ddd.adapters.job.memory_scheduler_adapter import MemorySchedulerAdapter
from ddd.domain.azure.azure_event_publisher import AzureEventPublisher
from ddd.domain.kafka.kafka_event_publisher import KafkaEventPublisher
from ddd.domain.memory.memory_event_publisher import MemoryEventPublisher
from ddd.infrastructure.job_service import JobService
from ddd.infrastructure.log_service import LogService
from ddd.infrastructure.memory.memory_db_service import MemoryDbService
from ddd.infrastructure.postgres.postgres_db_service import PostgresDbService
from ddd.repositories.memory.memory_dummy_repository import \
    MemoryDummyRepository
from ddd.repositories.memory.memory_event_repository import \
    MemoryEventRepository
from ddd.repositories.postgres.postgres_dummy_repository import \
    PostgresDummyRepository


class DependencyManager(object, metaclass=ABCMeta):
    """
    A dependency manager used to create & retrieve dependencies.
    """
    def __init__(
        self,
        config,
        loop=None,
    ):
        super().__init__()

        # Vars
        self.config = config
        self.loop = loop if loop else asyncio.get_event_loop()
        self.service = None

        # Dependencies
        self.db_service = None
        self.domain_adapter = None
        self.domain_publisher = None
        self.dummy_repository = None
        self.event_repository = None
        self.interchange_adapter = None
        self.interchange_publisher = None
        self.job_service = None
        self.job_adapter = None
        self.log_service = None
        self.scheduler_adapter = None

    def get_db_service(self):
        if not self.db_service:
            if self.config.database.type == "memory":
                self.db_service = \
                    MemoryDbService(
                        log_service=self.get_log_service(),
                    )
            elif self.config.database.type == "postgres":
                self.db_service = \
                    PostgresDbService(
                        dsn=self.config.database.postgres.dsn,
                        log_service=self.get_log_service(),
                        min_size=20,
                        max_size=20,
                    )
            else:
                raise Exception(
                    f"Db service for type not supported: "
                    f"'{self.config.database.type}'."
                )
        return self.db_service

    def get_domain_adapter(self):
        if not self.domain_adapter:
            if self.config.pubsub.domain.provider == "memory":
                self.domain_adapter = MemoryEventAdapter(
                    log_service=self.get_log_service(),
                )
            elif self.config.pubsub.domain.provider == "kafka":
                self.domain_adapter = KafkaEventAdapter(
                    bootstrap_servers=
                    self.config.pubsub.kafka.bootstrap_servers,
                    topic=self.config.pubsub.domain.topic,
                    group=self.config.pubsub.domain.group,
                    log_service=self.get_log_service(),
                )
            elif self.config.pubsub.domain.provider == "azure":
                self.domain_adapter = AzureEventAdapter(
                    namespace_conn_string=
                    self.config.pubsub.azure.namespace_conn_string,
                    checkpoint_store_conn_string=
                    self.config.pubsub.azure.checkpoint_store_conn_string,
                    blob_container_name=
                    self.config.pubsub.azure.blob_container_name,
                    topic=self.config.pubsub.domain.topic,
                    group=self.config.pubsub.domain.group,
                    log_service=self.get_log_service(),
                )
            else:
                raise Exception(
                    f"Domain adapter for type not supported: "
                    f"'{self.config.pubsub.domain.provider}'."
                )

        return self.domain_adapter

    def get_domain_publisher(self):
        if not self.domain_publisher:
            if self.config.pubsub.domain.provider == "memory":
                self.domain_publisher = MemoryEventPublisher(
                    log_service=self.get_log_service(),
                    keep_flushed_copies=True,
                )
                self.domain_publisher.set_primary_adapter(
                    primary_adapter=self.get_domain_adapter()
                )
            elif self.config.pubsub.domain.provider == "kafka":
                self.domain_publisher = KafkaEventPublisher(
                    bootstrap_servers=
                    self.config.pubsub.kafka.bootstrap_servers,
                    topic=self.config.pubsub.domain.topic,
                    group=self.config.pubsub.domain.group,
                    log_service=self.get_log_service(),
                    keep_flushed_copies=True,
                )
            elif self.config.pubsub.domain.provider == "azure":
                self.domain_publisher = AzureEventPublisher(
                    namespace=self.config.pubsub.azure.namespace,
                    namespace_conn_string=
                    self.config.pubsub.azure.namespace_conn_string,
                    topic=self.config.pubsub.domain.topic,
                    group=self.config.pubsub.domain.group,
                    log_service=self.get_log_service(),
                    keep_flushed_copies=True,
                )
            else:
                raise Exception(
                    f"Domain publisher for type not supported: "
                    f"'{self.config.pubsub.domain.provider}'."
                )
        return self.domain_publisher

    def get_dummy_repository(self):
        if not self.dummy_repository:
            db_type = self.config.database.type
            if db_type == "memory":
                self.dummy_repository = \
                    MemoryDummyRepository(
                        log_service=self.get_log_service(),
                    )
            elif db_type == "postgres":
                self.dummy_repository = \
                    PostgresDummyRepository(
                        config=self.config,
                        db_service=self.get_db_service(),
                        log_service=self.get_log_service(),
                    )
            else:
                raise Exception("Unsupported db type: '{db_type}'")
        return self.dummy_repository

    def get_event_repository(self):
        if not self.event_repository:
            self.event_repository = MemoryEventRepository()
        return self.event_repository

    def get_interchange_adapter(self):
        if not self.interchange_adapter:

            if self.config.pubsub.interchange.provider == "memory":
                self.interchange_adapter = MemoryEventAdapter(
                    log_service=self.get_log_service(),
                )
            elif self.config.pubsub.interchange.provider == "kafka":
                self.interchange_adapter = KafkaEventAdapter(
                    bootstrap_servers=
                    self.config.pubsub.kafka.bootstrap_servers,
                    topic=self.config.pubsub.interchange.topic,
                    group=self.config.pubsub.interchange.group,
                    log_service=self.get_log_service(),
                )
            elif self.config.pubsub.interchange.provider == "azure":
                self.interchange_adapter = AzureEventAdapter(
                    namespace_conn_string=
                    self.config.pubsub.azure.namespace_conn_string,
                    checkpoint_store_conn_string=
                    self.config.pubsub.azure.checkpoint_store_conn_string,
                    blob_container_name=
                    self.config.pubsub.azure.blob_container_name,
                    topic=self.config.pubsub.interchange.topic,
                    group=self.config.pubsub.interchange.group,
                    log_service=self.get_log_service(),
                )
            else:
                raise Exception(
                    f"Interchange adapter for type not supported: "
                    f"'{self.config.pubsub.interchange.provider}'."
                )

        return self.interchange_adapter

    def get_interchange_publisher(self):
        if not self.interchange_publisher:
            if self.config.pubsub.interchange.provider == "memory":
                self.interchange_publisher = MemoryEventPublisher(
                    log_service=self.get_log_service(),
                    keep_flushed_copies=True,
                )
                self.interchange_publisher.set_primary_adapter(
                    primary_adapter=self.get_interchange_adapter()
                )
            elif self.config.pubsub.interchange.provider == "kafka":
                self.interchange_publisher = KafkaEventPublisher(
                    bootstrap_servers=
                    self.config.pubsub.kafka.bootstrap_servers,
                    topic=self.config.pubsub.interchange.topic,
                    group=self.config.pubsub.interchange.group,
                    log_service=self.get_log_service(),
                    keep_flushed_copies=True,
                )
            elif self.config.pubsub.interchange.provider == "azure":
                self.interchange_publisher = AzureEventPublisher(
                    namespace=self.config.pubsub.azure.namespace,
                    namespace_conn_string=
                    self.config.pubsub.azure.namespace_conn_string,
                    topic=self.config.pubsub.interchange.topic,
                    group=self.config.pubsub.interchange.group,
                    log_service=self.get_log_service(),
                    keep_flushed_copies=True,
                )
            else:
                raise Exception(
                    f"Interchange publisher for type not supported: "
                    f"'{self.config.pubsub.interchange.provider}'."
                )
        return self.interchange_publisher

    def get_job_adapter(self):
        if not self.job_adapter:
            self.job_adapter = \
                JobAdapter(
                    log_service=self.get_log_service(),
                    token=self.config.auth.full_access_token,
                )
        return self.job_adapter

    def get_job_service(self):
        if not self.job_service:
            self.job_service = \
                JobService(
                    scheduler_adapter=self.get_scheduler_adapter(),
                    log_service=self.log_service,
                )
        return self.job_service

    def get_log_service(self):
        if not self.log_service:
            slack_config = self.config.slack
            slack_config.token = self.config.slack.token
            slack_config.bot_username = "Log Service (ddd-for-python)"

            kibana_config = self.config.log.kibana
            kibana_config.source = "Log Service (ddd-for-python)"

            self.log_service = \
                LogService(
                    kibana_config=kibana_config,
                    slack_config=slack_config,
                )

        return self.log_service

    def get_scheduler_adapter(self):
        if not self.scheduler_adapter:
            if self.config.jobs.scheduler.type == "apscheduler":
                self.scheduler_adapter = ApSchedulerAdapter(
                    dsn=self.config.jobs.scheduler.dsn,
                    exec_func=self.get_job_adapter().run_job,
                    log_service=self.get_log_service(),
                )
            elif self.config.jobs.scheduler.type == "memory":
                self.scheduler_adapter = MemorySchedulerAdapter(
                    exec_func=self.get_job_adapter().run_job,
                    log_service=self.get_log_service(),
                )
            else:
                raise Exception(
                    f"Scheduler adapter for type not supported: "
                    f"'{self.config.jobs.scheduler.type}'."
                )
        return self.scheduler_adapter

    def get_service(self):
        return self.service

    def set_service(self, service):
        self.service = service
        self.wire_together()

    def wire_together(self):
        self.get_domain_adapter().set_service(self.service)
        self.get_interchange_adapter().set_service(self.service)
        self.get_job_adapter().set_service(self.service)



from ddd.repositories.repository import Repository


class MemoryRepository(Repository):

    def __init__(
        self,
        log_service,
        aggregate_name,
        aggregate_cls,
        translator_cls,
    ):
        super().__init__(
            log_service=log_service,
            aggregate_cls=aggregate_cls,
            translator_cls=translator_cls,
        )

        self.highest_id = 0
        self.highest_id_collector = 0
        self.records = {}

        self.aggregate_name = aggregate_name

    async def next_identity(self):
        self.highest_id += 1
        return str(self.highest_id)

    async def next_collector_identity(self):
        self.highest_id_collector += 1
        return str(self.highest_id_collector)

    # Operations

    async def _get_record(self, aggregate_id):
        if str(aggregate_id) in self.records:
            return self.records[str(aggregate_id)]
        return None

    async def _get_all_records(self):
        return self.records.values()

    async def _get_all_records_not_on_latest_version(self):
        records = []

        for r in self.records.values():
            if int(r['data']['version']) < int(self.aggregate_cls.VERSION):
                records.append(r)

        return records

    async def _save_record(self, aggregate_id, data):
        aggregate_id = str(aggregate_id)

        if aggregate_id in self.records:
            self.records.pop(aggregate_id)

        self.records[aggregate_id] = {
            'id': aggregate_id,
            'data': data,
        }

    async def _delete(self, aggregate_id):
        self.records = {
            agg_id: data for agg_id, data in self.records.items()
            if agg_id != aggregate_id
        }

    async def get_count(self):
        return len(self.records)

    async def get_single_or_raise(self):
        """
        Convenience method for tests to get a single aggregate.
        Throws exception if not exactly 1 aggregate in repository.
        """
        aggregates = await self.get_all()

        if len(aggregates) != 1:
            raise Exception(
                "Couldn't get single aggregate because there are not "
                "exactly 1 aggregate in the repository."
            )

        return aggregates[0]

    async def get_all(self):
        records = await self._get_all_records()

        records = self._migrate(records, self.aggregate_cls)

        aggregates = [
            self.aggregate_from_record(record)
            for record in records
        ]

        return aggregates

    async def delete_all(self):
        self.highest_id = 0
        self.records = {}

    # Control

    async def start(self):
        pass

    async def stop(self):
        pass

    async def connect(self):
        pass

    async def assert_table(self):
        pass

    async def disconnect(self):
        pass

    # Clean

    async def empty(self):
        self.highest_id = 0
        self.records = {}



from ddd.tests.dummy_action_test_case import DummyActionTestCase
from ddd.domain.dummy.dummy import Dummy
from ddd.repositories.memory.memory_dummy_repository import \
    MemoryDummyRepository


class TestMemoryRepository(DummyActionTestCase):

    async def asyncSetUp(self):
        await super().asyncSetUp()

        self.repository = \
            MemoryDummyRepository(
                log_service=self.deps.get_log_service(),
            )

        Dummy.VERSION = "2"

        self.record_v1 = {
            'id': "some-dummy-id-1",
            'data': {
                'version': "1",
            }
        }

        self.record_v2 = {
            'id': "some-dummy-id-2",
            'data': {
                'version': "2",
            }
        }

    async def test_get_all_records_not_on_latest_version_returns_not_on_latest_version(
        self,
    ):
        # Setup
        await self.repository._save_record(
            aggregate_id=self.record_v1['id'],
            data=self.record_v1['data'],
        )

        await self.repository._save_record(
            aggregate_id=self.record_v2['id'],
            data=self.record_v2['data'],
        )

        Dummy.VERSION = "2"

        # Exercise
        records = await \
            self.repository._get_all_records_not_on_latest_version()

        # Assert
        self.assertEqual(
            [
                self.record_v1,
            ],
            records,
        )



from ddd.adapters.message_reader import MessageReader

from ddd.utils.utils import str_or_none


class DummyTranslator(object):
    def __init__(self, dummy):
        super().__init__()
        self.dummy = dummy

    def to_interchange(self):
        return self.to_dict()

    def to_domain(self):
        return self.to_dict()

    @classmethod
    def from_domain(cls, the_dict):
        from ddd.domain.dummy.dummy import Dummy
        from ddd.domain.dummy.dummy_id import DummyId

        reader = MessageReader(the_dict)

        # Read
        dummy = Dummy(
            version=reader.string_value('version'),
            dummy_id=reader.entity_id_value('dummy_id', DummyId),
            property_added=reader.string_value('property_added'),
        )

        return dummy

    def to_record(self):
        return self.to_dict()

    @classmethod
    def from_record(cls, record):
        the_dict = record['data']
        the_dict['dummy_id'] = record['id']

        return DummyTranslator.from_domain(the_dict)

    def to_dict(self):
        return {
            'version': self.dummy.version,
            'dummy_id': str_or_none(self.dummy.dummy_id),
            'property_added': self.dummy.property_added,
        }



from ddd.domain.event import Event


class DomainEvent(Event):
    """
    An domain event.
    """
    def __init__(self, name, version="1", date=None, sender=None, corr_ids=None):
        super(DomainEvent, self).__init__(
            name=name,
            version=version,
            date=date,
            sender=sender,
            corr_ids=corr_ids,
        )



import asyncio

from asyncio import Semaphore


class ApplicationService(object):
    """
    An application service.
    Extend this to add your own dependencies.
    """
    def __init__(
        self,
        db_service,
        domain_adapter,
        domain_publisher,
        event_repository,
        interchange_adapter,
        interchange_publisher,
        job_service,
        job_adapter,
        log_service,
        scheduler_adapter,
        max_concurrent_actions=40,
        loop=None,
    ):
        super().__init__()

        # Dependencies
        self.db_service = db_service
        self.domain_adapter = domain_adapter
        self.domain_publisher = domain_publisher
        self.event_repository = event_repository
        self.interchange_adapter = interchange_adapter
        self.interchange_publisher = interchange_publisher
        self.job_service = job_service
        self.job_adapter = job_adapter
        self.log_service = log_service
        self.scheduler_adapter = scheduler_adapter

        # ..categorize
        self.primary_adapters = [
            domain_adapter,
            interchange_adapter,
            job_adapter,
        ]

        self.secondary_adapters = []

        self.domain_services = []

        self.infrastructure_services = [
            db_service,
            job_service,
            log_service,
        ]

        # Vars
        self.loop = loop if loop else asyncio.get_event_loop()
        self.max_concurrent_actions = max_concurrent_actions

        # Actions
        self._create_action_sem()

    # Container

    def _create_action_sem(self):
        """
        Create the semaphore used to throttle concurrent actions.
        This number affects e.g. database connections opened.
        """
        self.action_sem = Semaphore(self.max_concurrent_actions)

    # Control

    async def start(self):
        """
        Starts the service.
        """
        self.log_service.info("Starting application service.")

        await self._start_domain_publisher()
        await self._start_interchange_publisher()
        await self._start_domain_services()
        await self._start_infrastructure_services()
        await self._start_secondary_adapters()
        await self._start_primary_adapters()

        self.log_service.info("Application service started.")

    async def stop(self):
        """
        Stops the service.
        """
        self.log_service.info("Stopping application service.")

        await self._stop_domain_publisher()
        await self._stop_interchange_publisher()
        await self._stop_primary_adapters()
        await self._stop_secondary_adapters()
        await self._stop_domain_services()
        await self._stop_infrastructure_services()

        self.log_service.info("Stopping application service.")

    async def _start_domain_publisher(self):
        """
        Starts the domain publisher.
        """
        self.log_service.info("Starting domain publisher.")

        await self.domain_publisher.start()

    async def _start_interchange_publisher(self):
        """
        Starts the interchange publisher.
        """
        self.log_service.info("Starting interchange publisher.")

        await self.interchange_publisher.start()

    async def _start_primary_adapters(self):
        """
        Starts the primary adapters.
        """
        self.log_service.info("Starting primary adapters..")

        for adapter in self.primary_adapters:
            if hasattr(adapter, "start"):
                self.log_service.info(
                    f"..starting {adapter.__class__.__name__}"
                )
                await adapter.start()

    async def _start_secondary_adapters(self):
        """
        Starts the secondary adapters.
        """
        self.log_service.info("Starting secondary adapters..")

        for adapter in self.secondary_adapters:
            if hasattr(adapter, "start"):
                self.log_service.info(
                    f"..starting {adapter.__class__.__name__}"
                )
                await adapter.start()

    async def _stop_interchange_publisher(self):
        """
        Stop the interchange publisher.
        """
        self.log_service.info("Stopping interchange publisher.")

        await self.interchange_publisher.stop()

    async def _stop_domain_publisher(self):
        """
        Stop the domain publisher.
        """
        self.log_service.info("Stopping domain publisher.")

        await self.domain_publisher.stop()

    async def _stop_primary_adapters(self):
        """
        Stops the primary adapters.
        """
        self.log_service.info("Stopping primary adapters..")

        for adapter in self.primary_adapters:
            if hasattr(adapter, "stop"):
                self.log_service.info(
                    f"..stopping {adapter.__class__.__name__}."
                )
                await adapter.stop()

    async def _stop_secondary_adapters(self):
        """
        Stops the secondary adapters.
        """
        self.log_service.info("Stopping secondary adapters..")

        for adapter in self.secondary_adapters:
            if hasattr(adapter, "stop"):
                self.log_service.info(
                    f"..stopping {adapter.__class__.__name__}."
                )
                await adapter.stop()

    async def _start_domain_services(self):
        """
        Starts the domain services.
        """
        self.log_service.info("Starting domain services..")

        for service in self.domain_services:
            if hasattr(service, "start"):
                self.log_service.info(
                    f"..starting {service.__class__.__name__}"
                )
                await service.start()

    async def _start_infrastructure_services(self):
        """
        Starts the infrastructure services.
        """
        self.log_service.info("Starting infrastructure services..")

        for service in self.infrastructure_services:
            if hasattr(service, "start"):
                self.log_service.info(
                    f"..starting {service.__class__.__name__}"
                )
                await service.start()

    async def _stop_domain_services(self):
        """
        Stops the domain services.
        """
        self.log_service.info("Stopping domain services..")

        for service in self.domain_services:
            if hasattr(service, "stop"):
                self.log_service.info(
                    f"..stopping {service.__class__.__name__}"
                )
                await service.stop()

    async def _stop_infrastructure_services(self):
        """
        Stops the infrastructure services.
        """
        self.log_service.info("Stopping infrastructure services..")

        for service in self.infrastructure_services:
            if hasattr(service, "stop"):
                self.log_service.info(
                    f"..stopping {service.__class__.__name__}"
                )
                await service.stop()



import asyncio
import uuid

from abc import ABCMeta

from ddd.repositories.repository import Repository


class PostgresRepository(Repository, metaclass=ABCMeta):

    def __init__(
        self,
        db_service,
        log_service,
        aggregate_cls,
        translator_cls,
        table_name,
        dsn,
        loop=None,
    ):
        super().__init__(
            log_service=log_service,
            aggregate_cls=aggregate_cls,
            translator_cls=translator_cls,
        )

        self.db_service = db_service
        self.table_name = table_name
        self.dsn = dsn
        self.loop = loop if loop is not None else asyncio.get_event_loop()

    # Operations

    async def _delete(self, aggregate_id):
        async with self.db_service.conn_pool.acquire() as conn:
            await conn.execute(
                """
                DELETE FROM {} WHERE id = $1
                """.
                format(
                    self.table_name
                ),
                str(aggregate_id)
            )

    async def _get_record(self, aggregate_id):
        async with self.db_service.conn_pool.acquire() as conn:
            record = await conn.fetchrow(
                """
                SELECT * FROM {} WHERE id = $1
                """.
                format(
                    self.table_name
                ),
                str(aggregate_id)
            )

            return record

    async def _get_all_records(self):
        async with self.db_service.conn_pool.acquire() as conn:
            records = await conn.fetch(
                """
                SELECT * FROM {}
                """.
                format(
                    self.table_name
                )
            )
            return records

    async def _get_all_records_not_on_latest_version(self):
        async with self.db_service.conn_pool.acquire() as conn:
            records = await conn.fetch(
                """
                SELECT * FROM {} WHERE (data->>'version')::int < $1
                """.
                format(
                    self.table_name
                ),
                int(self.aggregate_cls.VERSION)
            )
            return records

    async def delete_all(self):
        await self.assert_table()

        async with self.db_service.conn_pool.acquire() as conn:
            await conn.execute(
                """
                DELETE FROM {}
                """.
                format(
                    self.table_name
                )
            )
            return True

    async def get_single_or_raise(self):
        """
        Convenience method for tests to get a single aggregate.
        Throws exception if not exactly 1 aggregate in repository.
        """
        async with self.db_service.conn_pool.acquire() as conn:
            records = await conn.fetch(
                """
                SELECT * FROM {}
                """.
                format(
                    self.table_name
                )
            )
            aggregates = [
                self.aggregate_from_record(record) for record in records
            ]

            if len(aggregates) != 1:
                raise Exception(
                    "Couldn't get single aggregate because there are not "
                    "exactly 1 aggregate in the repository."
                )

            return aggregates[0]

    async def get_count(self):
        """
        Get the total number of aggregates.
        """
        async with self.db_service.conn_pool.acquire() as conn:
            count = await conn.fetchval(
                """
                SELECT COUNT(*) FROM {}
                """.
                format(
                    self.table_name
                )
            )
            return count

    async def next_identity(self):
        return uuid.uuid4().hex

    async def _save_record(self, aggregate_id, data):
        async with self.db_service.conn_pool.acquire() as conn:
            await conn.execute(
                """
                INSERT INTO {} (id, data)
                VALUES
                (
                    $1,
                    $2
                ) 
                ON CONFLICT (id)
                DO
                UPDATE
                    SET data = $2
                """.
                format(
                    self.table_name
                ),
                str(aggregate_id),
                data,
            )

    def _filter_by_property(self, records, property, values):
        return \
            self._filter_by_properties(
                records=records,
                properties={
                    property: values,
                },
            )

    def _filter_by_properties(self, records, properties):
        filtered = []

        for record in records:
            for name, values in properties.items():
                if name not in record['data']:
                    raise Exception(
                        f"Couldn't filter records by property values in "
                        f"postgres adapter. The record has no property "
                        f"named '{name}'."
                    )
                if record['data'][name] in values:
                    filtered.append(record)

        return filtered

    # Connection

    async def connect(self):
        pass

    async def disconnect(self):
        pass

    # Table

    async def assert_table(self):
        async with self.db_service.conn_pool.acquire() as conn:
            await conn.execute(
                """
                CREATE TABLE IF NOT EXISTS {}(
                    id VARCHAR UNIQUE NOT NULL,
                    data json NOT NULL
                )
                """.
                format(
                    self.table_name
                )
            )



class BuildingBlock(object):
    def __init__(self, version):
        super(BuildingBlock, self).__init__()
        self.version = version

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            a = self.__dict__
            b = other.__dict__
            return a == b
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def equals(self, other, ignore_fields):
        ignore_fields = ignore_fields if ignore_fields is not None else []

        if isinstance(other, self.__class__):

            from ddd.utils.utils import get_for_compare

            a = get_for_compare(self, ignore_fields)
            b = get_for_compare(other, ignore_fields)

            return a == b
        else:
            return False

    def serialize(self, ignore_fields=None):
        """
        Serializes into a dict.
        """
        from ddd.utils.utils import get_for_compare
        return get_for_compare(self, ignore_fields=ignore_fields)



from abc import ABCMeta

from ddd.adapters.message_reader import MessageReader


class EventListener(object, metaclass=ABCMeta):
    """
    An event listener.
    """
    def __init__(
        self,
        listens_to,
        action,
        command_creator,
        service=None,
        enabled=True,
    ):
        super(EventListener, self).__init__()

        self.listens_to_name = listens_to
        self.action = action
        self.service = service
        self.command_creator = command_creator
        self.enabled = enabled

    def listens_to(self, name):
        """
        Used to check if we listen to an event named 'name'.
        """
        if not self.listens_to_name:
            raise Exception("The 'listens_to_name' variable hasn't been set.")
        return name == self.listens_to_name

    async def handle(self, message):
        """
        Called when an event we are listening to is received.
        """
        if not self.enabled:
            return

        if self.command_creator is None:
            raise Exception(
                "Can't delegate event because command_creator is null.")

        self.message = message
        self.event = self.read_event()

        try:
            command = \
                self.command_creator(
                    event=self.event
                )
        except Exception as e:
            raise Exception(
                f"Couldn't create command using command "
                f"creator, error: {str(e)}"
            )
        action = getattr(self.service, self.action)

        await action(
            command=command,
            corr_ids=self.event.corr_ids,
        )

    def read_event(self):
        """
        Reads basic information of the event.
        """
        # Create reader
        self.reader = MessageReader(self.message)

        # Common properties
        self.type = self.reader.string_value('type')
        self.version = self.reader.string_value('version')
        self.name = self.reader.string_value('name')
        self.action_id = self.reader.string_value('action_id')
        self.corr_ids = self.reader.list_value('corr_ids')

    def enable(self):
        self.enabled = True

    def disable(self):
        self.enabled = False
